"""
TODO:
 - https://github.com/aristocratos/btop/blob/00c90884c328eb3e44a0ab094e833161092aae9c/src/linux/btop_collect.cpp#L443

==

Repls
 - python - https://docs.python.org/3/tutorial/interpreter.html
 - ipython - https://github.com/ipython/ipython/
 - ptpython - https://github.com/prompt-toolkit/ptpython
 - bpython - https://github.com/bpython/bpython

Debuggers
 - pdb - https://docs.python.org/3/library/pdb.html
 - ipdb - https://github.com/gotcha/ipdb
 - pdub - https://github.com/inducer/pudb/

CPU Profilers
 - cProfile - https://docs.python.org/3/library/profile.html
 - pyinstrument - https://github.com/joerick/pyinstrument
 - py-pspy - https://github.com/benfred/py-spy
 - austin-dist - https://github.com/P403n1x87/austin
 - yappi - https://github.com/sumerc/yappi

Memory Profilers
 - tracemalloc - https://docs.python.org/3/library/tracemalloc.html
 - guppy3 - https://github.com/zhuyifei1999/guppy3/
 - objgraph - https://github.com/mgedmin/objgraph
 - memory-profiler - https://github.com/pythonprofilers/memory_profiler
 - memray - https://github.com/bloomberg/memray
 - pympler - https://github.com/pympler/pympler
"""
